local date = require("date")
local socketurl = require("socket.url")
local utils = require("escher.utils")
local Canonicalizer = require("escher.canonicalizer")
local Signer = require("escher.signer")

local Escher = {}

local meta = {
  __index = Escher
}

function Escher:new(options)
  options = options or {}

  local object = {
    algoPrefix = options.algoPrefix or "ESR",
    vendorKey = options.vendorKey or "ESCHER",
    hashAlgo = options.hashAlgo or "SHA256",
    apiSecret = options.apiSecret,
    accessKeyId = options.accessKeyId,
    credentialScope = options.credentialScope or "escher_request",
    authHeaderName = options.authHeaderName or "X-Escher-Auth",
    dateHeaderName = options.dateHeaderName or "X-Escher-Date",
    date = date(options.date or os.date("!%c")),
    clockSkew = options.clockSkew or 300
  }

  return setmetatable(object, meta)
end

setmetatable(Escher, {
  __call = Escher.new
})

local function getHeaderValue(headers, headerName)
  for _, header in ipairs(headers) do
    local name = utils.trim(header[1]:lower())

    if name == headerName:lower() then
      return header[2]
    end
  end
end

local function throwError(error)
  return false, error
end

local function parseQuery(query, algoPrefix)
  local hashAlgo = string.match(query, algoPrefix .. "%-HMAC%-(%w+)&")
  local accessKeyId, shortDate, credentialScope = string.match(query, "Credentials=([A-Za-z0-9%-%_]+)/(%d+)/([A-Za-z0-9%-%_%/]-)&")
  local signedHeaders = string.match(query, "SignedHeaders=([a-z0-9%-%_%;]+)&")
  local signature = string.match(query, "Signature=([a-f0-9]+)")

  return {
    hashAlgo = hashAlgo,
    accessKeyId = accessKeyId,
    shortDate = shortDate,
    credentialScope = credentialScope,
    signedHeaders = signedHeaders,
    signature = signature
  }
end

local function stripSignatureFromQuery(url)
  local strippedUrl = ""

  for _, value in ipairs(utils.split(url, "&")) do
    if not string.match(value, "Signature") then
      strippedUrl = strippedUrl .. value .. "&"
    end
  end

  return string.sub(strippedUrl, 1, -2)
end

local function parseAuthHeader(authHeader, algoPrefix)
  local hashAlgo, accessKeyId, shortDate, credentialScope, signedHeaders, signature = string.match(
    authHeader,
    algoPrefix ..
    "%-HMAC%-(%w+)%s+" ..
    "Credential=([A-Za-z0-9%-%_]+)/(%d+)/([A-Za-z0-9%-%_%/% ]-),%s*" ..
    "SignedHeaders=([a-zA-Z0-9%-%_%;]+),%s*" ..
    "Signature=([a-f0-9]+)"
  )

  return {
    hashAlgo = hashAlgo,
    accessKeyId = accessKeyId,
    shortDate = shortDate,
    credentialScope = credentialScope,
    signedHeaders = signedHeaders,
    signature = signature
  }
end

local function isDateWithinRange(requestDate, signedDate, expires)
  local diff = math.abs(date.diff(requestDate, signedDate):spanseconds())

  return diff <= expires
end

function Escher:authenticate(request, getApiSecret, mandatorySignedHeaders)
  request = utils.merge({}, request)

  local uri = socketurl.parse(request.url)
  local isPresignedUrl = string.match(uri.query or "", "Signature") and request.method == "GET"

  local dateHeader = getHeaderValue(request.headers, self.dateHeaderName)
  local authHeader = getHeaderValue(request.headers, self.authHeaderName)
  local hostHeader = getHeaderValue(request.headers, "host")

  if not dateHeader and not isPresignedUrl then
    return throwError("The " .. self.dateHeaderName:lower() .. " header is missing")
  end

  if not authHeader and not isPresignedUrl then
    return throwError("The " .. self.authHeaderName:lower() .. " header is missing")
  end

  if not hostHeader then
    return throwError("The host header is missing")
  end

  local authParts
  local expires
  local requestDate

  if isPresignedUrl then
    requestDate = date(string.match(uri.query, "Date=([A-Za-z0-9]+)&"))
    authParts = parseQuery(socketurl.unescape(uri.query), self.algoPrefix)
    expires = tonumber(string.match(uri.query, "Expires=([0-9]+)&"))
    request.url = stripSignatureFromQuery(request.url)
    request.body = "UNSIGNED-PAYLOAD"
  else
    requestDate = date(dateHeader)
    authParts = parseAuthHeader(authHeader or "", self.algoPrefix)
    expires = 0
  end

  if not authParts.hashAlgo then
    return throwError("Could not parse " .. self.authHeaderName .. " header")
  end

  local headersToSign = utils.split(authParts.signedHeaders, ";")

  for _, header in ipairs(headersToSign) do
    if string.lower(header) ~= header then
      return throwError("SignedHeaders must contain lowercase header names in the " .. self.authHeaderName .. " header")
    end
  end

  if type(mandatorySignedHeaders) ~= "table" and mandatorySignedHeaders ~= nil then
    return throwError("The mandatorySignedHeaders parameter must be undefined or array of strings")
  end

  mandatorySignedHeaders = utils.merge({}, mandatorySignedHeaders)

  table.insert(mandatorySignedHeaders, "host")

  if not isPresignedUrl then
    table.insert(mandatorySignedHeaders, self.dateHeaderName:lower())
  end

  for _, header in ipairs(mandatorySignedHeaders) do
    if type(header) ~= "string" then
      return throwError("The mandatorySignedHeaders parameter must be undefined or array of strings")
    end

    if not utils.contains(headersToSign, header) then
      return throwError("The " ..  header .. " header is not signed")
    end
  end

  if authParts.credentialScope ~= self.credentialScope then
    return throwError("The credential scope is invalid")
  end

  if authParts.hashAlgo ~= self.hashAlgo then
    return throwError("Only SHA256 and SHA512 hash algorithms are allowed")
  end

  if authParts.shortDate ~= utils.toShortDate(requestDate) then
    return throwError("The " .. self.authHeaderName .. " header's shortDate does not match with the request date")
  end

  if not isDateWithinRange(self.date, requestDate, self.clockSkew + expires) then
    return throwError("The request date is not within the accepted time range")
  end

  local apiSecret = getApiSecret(authParts.accessKeyId)

  if apiSecret == nil then
    return throwError("Invalid Escher key")
  end

  if authParts.signature ~= Signer(self):calculateSignature(request, headersToSign, date(requestDate), apiSecret) then
    return throwError("The signatures do not match")
  end

  return authParts.accessKeyId
end

local function hasHeader(headers, headerName)
  for _, values in ipairs(headers) do
    if values[1]:lower() == headerName:lower() then
      return true
    end
  end

  return false
end

local function addDateHeaderIfNotExists(self, headers)
  if not hasHeader(headers, self.dateHeaderName) then
    table.insert(headers, { self.dateHeaderName, self.date:fmt("${http}") })
  end
end

local function generateFullCredentials(self)
  return string.format("%s/%s/%s", self.accessKeyId, utils.toShortDate(self.date), self.credentialScope)
end

function Escher:generateHeader(request, headersToSign)
  request = utils.merge({}, request)
  request.headers = utils.merge({}, request.headers)

  addDateHeaderIfNotExists(self, request.headers)

  return self.algoPrefix .. "-HMAC-" .. self.hashAlgo ..
    " Credential=" .. generateFullCredentials(self) ..
    ", SignedHeaders=" .. Canonicalizer(self):canonicalizeSignedHeaders(request.headers, headersToSign) ..
    ", Signature=" .. Signer(self):calculateSignature(request, headersToSign, self.date, self.apiSecret)
end

function Escher:signRequest(request, headersToSign)
  local authHeader = self:generateHeader(request, headersToSign)

  addDateHeaderIfNotExists(self, request.headers)

  table.insert(request.headers, { self.authHeaderName, authHeader })
end

local ONE_DAY_IN_SECONDS = 86400

local function getHash(parsedUrl)
  if parsedUrl.fragment then
    return "#" .. parsedUrl.fragment
  end

  return ""
end

local function stripHash(parsedUrl)
  parsedUrl = utils.merge({}, parsedUrl)

  parsedUrl.fragment = ""

  local builtUrl = socketurl.build(parsedUrl)

  return string.gsub(builtUrl, "#", "")
end

local function encodeSlashes(str)
  return string.gsub(str, "/", "%%2F")
end

function Escher:generatePreSignedUrl(url, client, expires)
  expires = expires or ONE_DAY_IN_SECONDS

  local headersToSign = { "host" }
  local parsedUrl = socketurl.parse(socketurl.unescape(url))
  local hash = getHash(parsedUrl)

  if #hash > 0 then
    url = stripHash(parsedUrl)
  end

  local signedUrl = table.concat({
    url,
    "X-EMS-Algorithm=" .. self.algoPrefix .. "-HMAC-" .. self.hashAlgo,
    "X-EMS-Credentials=" .. encodeSlashes(client[1] .. "/" .. utils.toShortDate(self.date) .. "/" .. self.credentialScope),
    "X-EMS-Date=" .. utils.toLongDate(self.date),
    "X-EMS-Expires=" .. expires,
    "X-EMS-SignedHeaders=" .. table.concat(headersToSign, ";")
  }, "&") .. "&"

  local parsedSignedUrl = socketurl.parse(signedUrl)

  local host = parsedUrl.host
  local path, query = parsedSignedUrl.path, parsedSignedUrl.query

  local request = {
    host = host,
    method = "GET",
    url = path .. "?" .. query,
    headers = {
      { "host", host }
    },
    body = "UNSIGNED-PAYLOAD"
  }

  local signature = Signer(self):calculateSignature(request, headersToSign, self.date, self.apiSecret)

  return signedUrl .. "X-EMS-Signature=" .. signature  .. hash
end

return Escher
